/* $Id: srvinfo.C,v 1.22 2000/01/05 07:29:54 dm Exp $ */

/*
 *
 * Copyright (C) 1998, 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfscd.h"
#include "rxx.h"

#if FIX_MNTPOINT
#define MPDOT "."
#endif /* FIX_MNTPOINT */

static const u_int64_t badfsid = (u_int64_t) -1;
static ihash<const str, srvinfo, &srvinfo::path, &srvinfo::hlink> srvtab;

static inline str
mpof (str fs)
{
#if FIX_MNTPOINT
  if (opt_fix_mntpoint)
    return strbuf ("%s/" MPDOT, sfsroot) << fs << "/r";
  else
#endif /* FIX_MNTPOINT */
    return strbuf ("%s/", sfsroot) << fs;
}

void
update_devdb (cbv cb)
{
  afs_root->unlink (".devdb");
  afs_root->link (afsreg::alloc (srvinfo::devlist ()), ".devdb");
  (*cb) ();
}

inline
srvinfo::srvinfo (const str &p, const str &hn, const sfs_hash &hid)
  : tmo (NULL), waiting (true), destroyed (false), cdmounted (false),
    visible_flag (true), devno (badfsid), error (0), path (p),
    hostname (hn), hostid (hid), cdp (NULL)
{
  assert (!srvtab[p]);
  srvtab.insert (this);
}

inline
srvinfo::~srvinfo ()
{
  timecb_remove (tmo);
  afs_sfsroot->unlink (path);
#if FIX_MNTPOINT
  afs_root->unlink (MPDOT << path);
#endif /* FIX_MNTPOINT */

  srvtab.remove (this);
  if (cdp) {
    cdp->servers.remove (this);
    if (cdp->c) {
      nfspath3 p (path);
      cdp->c->call (SFSCDPROC_UNMOUNT, &p, NULL, aclnt_cb_null);
    }
  }
  if (waitq.empty ())
    flushpath (path);
  else
    for (size_t n = waitq.size (); n-- > 0;)
      (*waitq.pop_front ()) (NULL, error ? error : EAGAIN);
}

void
srvinfo::timeout (bool start)
{
  tmo = NULL;
  if (timecb_t *ntmo = timecb (time (NULL) + (destroyed ? 5 : 300),
			       wrap (this, &srvinfo::timeout, false))) {
    tmo = ntmo;
    if (!start)
      unmount ((destroyed ? NUOPT_FORCE : 0) | NUOPT_NLOG);
  }
}

void
srvinfo::fail (int err)
{
  error = err;
  delete this;
}

srvinfo *
srvinfo::lookup (const str &path)
{
  srvinfo *si = srvtab[path];
  if (si && !si->waiting)
    return si;
  return NULL;
}

void
srvinfo::alloc (const str &path, srvinfo::alloccb_t cb)
{
  srvinfo *si = srvtab[path];
  if (si) {
    if (si->waiting)
      si->waitq.push_back (cb);
    else if (si->error)
      (*cb) (NULL, si->error);
    else
      (*cb) (si, 0);
    return;
  }

  str hname;
  sfs_hash hid;
  if (!sfs_parsepath (path, &hname, &hid)) {
    (*cb) (NULL, ENOENT);
    return;
  }

  si = New srvinfo (path, hname, hid);
  si->waitq.push_back (cb);
  tcpconnect (si->hostname, sfs_port, wrap (si, &srvinfo::connected));
}

void
srvinfo::connected (int fd)
{
  if (fd < 0) {
    warn ("%s: %m\n", path.cstr ());
    fail (errno);
    return;
  }
  else {
    sockaddr_in sin;
    socklen_t len = sizeof (sin);
    bzero (&sin, sizeof (sin));
    if (getpeername (fd, reinterpret_cast<sockaddr *> (&sin), &len) < 0) {
      close (fd);
      warn ("%s: %m\n", path.cstr ());
      fail (errno);
      return;
    }
    if (badaddrs[sin.sin_addr]) {
      close (fd);
      warn ("%s: cannot connect to my own IP address\n", path.cstr ());
      fail (EDEADLK);
      return;
    }
  }

  tcp_nodelay (fd);
  ptr<aclnt> c = aclnt::alloc (axprt_stream::alloc (fd), sfs_program_1);
  sfs_connectarg arg;
  arg.release = sfs_release;
  arg.service = SFS_SFS;
  arg.name = hostname;
  arg.hostid = hostid;
  arg.extensions.set (sfs_extensions.base (), sfs_extensions.size (),
		      freemode::NOFREE);
  c->call (SFSPROC_CONNECT, &arg, &conres,
	   wrap (this, &srvinfo::gotconres, fd));
}

void
srvinfo::gotconres (int fd, clnt_stat err)
{
  if (err) {
    warn << path << ": " << err << "\n";
    fail (EIO);
    return;
  }
  if (conres.status) {
    warn << path << ": " << conres.status << "\n";
    switch (conres.status) {
    default:
    case SFS_NOSUCHHOST:
      fail (ENOENT);
      return;
    case SFS_NOTSUPP:
      fail (EPROTONOSUPPORT);
      return;
    case SFS_TEMPERR:
      fail (EAGAIN);
      return;
    case SFS_REDIRECT:
      {
	ptr<revocation> r = revocation::alloc (*conres.revoke);
	if (r) {
	  timecb (timenow + 300, wrap (r, &revocation::nop));
	  flushpath (path);
	  for (size_t n = waitq.size (); n-- > 0;)
	    (*waitq.pop_front ()) (NULL, 0);
	}
	fail (EAGAIN);
	return;
      }
    }
  }

  if (conres.reply->servinfo.host.hostname != hostname) {
    warn << path << ": server is '"
	 << conres.reply->servinfo.host.hostname << "'\n";
    fail (ENOENT);
    return;
  }
  if (!sfs_ckhostid (&hostid, conres.reply->servinfo.host)) {
    warn << path << ": hostid mismatch\n";
    fail (ENOENT);
    return;
  }
  
  cdp = release::cdlookup (conres.reply->servinfo.release,
			   conres.reply->servinfo.prog,
			   conres.reply->servinfo.vers);
  if (!cdp) {
    warn << path << ": no client for program "
	 << conres.reply->servinfo.prog << ", version "
	 << conres.reply->servinfo.vers
	 << " (release " << conres.reply->servinfo.release / 100
	 << "." << conres.reply->servinfo.release % 100
	 << ")\n";
    fail (EPROTONOSUPPORT);
    return;
  }
  cdp->servers.insert_head (this);

  if (!cdp->c) {
    warn << path << ": client for program "
	 << conres.reply->servinfo.prog << ", version "
	 << conres.reply->servinfo.vers << " is dead\n";
    fail (EAGAIN);
    return;
  }
  cdp->x->sendfd (dup (fd));

  if (!cdp->c) {
    warn << path << ": client for program "
	 << conres.reply->servinfo.prog << ", version "
	 << conres.reply->servinfo.vers << " is dead\n";
    fail (EAGAIN);
    return;
  }

  sfscd_mountarg arg;
  arg.carg.release = sfs_release;
  arg.carg.service = SFS_SFS;
  arg.carg.name = hostname;
  arg.carg.hostid = hostid;
  arg.carg.extensions.set (sfs_extensions.base (), sfs_extensions.size (),
			   freemode::NOFREE);
  arg.cres = *conres.reply;
  cdp->c->call (SFSCDPROC_MOUNT, &arg, &mntres,
		wrap (this, &srvinfo::gotmntres));
}

void
srvinfo::gotmntres (clnt_stat err)
{
  if (err || mntres.err) {
    fail (EIO);
    return;
  }
  cdmounted = true;
  int fd = cdp->x->recvfd ();
  if (fd < 0) {
    warn << "srvinfo::gotmntres: did not receive file descriptor\n";
    fail (EIO);
    return;
  }

#if FIX_MNTPOINT
  if (opt_fix_mntpoint) {
    str dpath (MPDOT << path);
    afs_root->mkdir (dpath)->mkdir ("r");
    afs_sfsroot->symlink (dpath << "/r", path);
  }
  else
#endif /* FIX_MNTPOINT */
    afs_sfsroot->mkdir (path);
  mnt_mount (fd, hostname, mpof (path),
	     mntres.reply->mntflags, mntres.reply->fh,
	     wrap (this, &srvinfo::gotnfsmntres));
}

void
srvinfo::gotnfsmntres (int err, u_int64_t fsid)
{
  if (err) {
    warn << "mount (" << path << "): " << strerror (errno) << "\n";
    fail (err);
    return;
  }
  devno = fsid;
  update_devdb (wrap (this, &srvinfo::ready));
}

void
srvinfo::ready ()
{
  while (!waitq.empty ())
    (*waitq.pop_front ()) (this, error);
  waiting = false;
  if (destroyed)
    unmount (NUOPT_FORCE);
  else
    timeout ();
}

void
srvinfo::unmountcb (cbi::ptr cb, int err)
{
  if (terminating && err != EBUSY && err != EAGAIN)
    err = 0;
  if (err)
    waiting = false;
  if (cb)
    (*cb) (err);
  if (!err) {
    if (waitq.empty ())
      delete this;
    else {
      vector <alloccb_t> wq;
      str p = path;
      swap (wq, waitq);
      delete this;
      while (!wq.empty ())
	srvinfo::alloc (p, wq.pop_front ());
    }
  }
}

void
srvinfo::unmount (int flags, cbi::ptr cb)
{
#if 0
  warn << "unmounting: " << path << " (flags = "
       << strbuf ("0x%x", flags) << ")\n";
#endif
  if (!waiting) {
    waiting = true;
    mnt_umount (mpof (path), flags, wrap (this, &srvinfo::unmountcb, cb));
  }
  else if (cb)
    (*cb) (EBUSY);
}

void
srvinfo::destroy (bool stale)
{
  if (waiting && stale) {
    destroyed = true;
    mnt_umount (mpof (path), NUOPT_STALE|NUOPT_NOOP, cbi_null);
  }
  else if (!destroyed) {
    timecb_remove (tmo);
    tmo = NULL;
    destroyed = true;
    unmount (NUOPT_FORCE | (stale ? NUOPT_STALE : 0));
    timeout ();
  }
}

void
srvinfo::destroy (const str &path, cdaemon *rcdp, bool stale)
{
  srvinfo *si = srvtab[path];
  if (si && (!rcdp || rcdp == si->cdp))
    si->destroy (stale);
}

void
srvinfo::show (const str &path, cdaemon *rcdp, bool showit)
{
  srvinfo *si = srvtab[path];
  if (si && (!rcdp || rcdp == si->cdp)) {
    si->visible_flag = showit;
    if (!showit && !si->waiting)
      si->unmount (NUOPT_NLOG);
  }
}

void
srvinfo::revoke (const str &path)
{
  srvinfo *si = srvtab[path];
  if (si) {
    if (si->cdp && si->cdp->c)
      si->cdp->c->call (SFSCDPROC_CONDEMN, &si->path, NULL, aclnt_cb_null);
    si->destroy (false);
  }
}

int
srvinfo::pathstate (const str &path)
{
  srvinfo *si = srvtab[path];
  if (!si)
    return 0;
  if (si->error)
    return si->error;
  if (si->waiting)
    return -1;
  if (!si->visible_flag)
    return EWOULDBLOCK;
  return 0;
}

void
srvinfo::printdev (strbuf *sb, srvinfo *si)
{
  if (si->devno != badfsid)
    sb->fmt ("0x%" U64F "x ", si->devno)
      << si->path << " " << si->cdp->name << "\n";
}

str
srvinfo::devlist ()
{
  strbuf sb;
  srvtab.traverse (wrap (*srvinfo::printdev, &sb));
  return sb;
}
