/* $Id: server.C,v 1.68 2000/01/08 19:44:23 dm Exp $ */

/*
 *
 * Copyright (C) 1998, 1999 David Mazieres (dm@uun.org)
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsrwcd.h"
#include "axprt_crypt.h"

/* #define NO_ACACHE 1 */

extern time_t timenow;

#define srvno_min 2U

u_int32_t server::srvno_next;
AUTH *aui0 = authuint_create (0);

//ihash<const sfs_hash, server, &server::hostid, &server::hidlink> &hostidtab
//  = *New ihash<const sfs_hash, server, &server::hostid, &server::hidlink>;

ihash<const str, server, &server::path, &server::pathlink> &pathtab
  = *New ihash<const str, server, &server::path, &server::pathlink>;
ihash<const u_int32_t, server, &server::srvno, &server::srvnolink> &srvnotab
  = *New ihash<const u_int32_t, server, &server::srvno, &server::srvnolink>;

static int nfsfd = -1;
#ifdef USE_UVFS
bool uvfs;
static int uvfsfd = -1;
static ptr<aclnt> uvfsclnt;
static ptr<asrv> uvfssrv;
static uvfsstate ustate;
#endif /* USE_UVFS */
static ptr<asrv> nfssrv;
blowfish fhkey;

static void
error_reply (svccb *sbp, nfsstat3 err)
{
#ifdef USE_UVFS
  if (uvfs)
    switch (err) {
    case NFS3ERR_BADHANDLE:
      uvfs_err (sbp, ESTALE);
      break;
    default:
      uvfs_err (sbp, err);
      break;
    }
  else
#endif /* USE_UVFS */
    nfs3_err (sbp, err);
}


void
server::init ()
{
#ifdef USE_UVFS
  uvfsfd = uvfs_open ();
  if (uvfsfd >= 0) {
    warn ("using uvfs\n");
    ref<axprt> uvfsx = axprt_stream::alloc (uvfsfd);
    uvfssrv = asrv::alloc (uvfsx, uvfsprog_1, wrap (&server::getcall));
    uvfsclnt = aclnt::alloc (uvfsx, uvfscbprog_1);
    if (!uvfssrv)
      fatal ("server_init: could not allocate uvfsprog_1 server\n");
    uvfs = true;
    return;
  }
#endif /* USE_UVFS */
  if (nfsfd != -1)
    panic ("server::init called twice\n");
  nfsfd = inetsocket (SOCK_DGRAM, 0, INADDR_LOOPBACK);
  if (nfsfd < 0)
    fatal ("socket: %m\n");
  ref<axprt> x = axprt_dgram::alloc (nfsfd);
  nfssrv = asrv::alloc (x, nfs_program_3, wrap (&server::getcall));
  if (!nfssrv)
    fatal ("server_init: could not allocate nfs_program_3 server\n");
}

void
server::getcall (svccb *sbp)
{
  if (sbp->proc () == NFSPROC3_NULL) {
    sbp->reply (NULL);
    return;
  }

  ptr<server> sp;
  u_int32_t srvno;
  if (!server::getserver (&sp, &srvno, sbp))
    error_reply (sbp, NFS3ERR_BADHANDLE);
  else if (sp)
    sp->dispatch (sbp);
  else if (!sfsctl_intercept (NULL, srvno, sbp))
    error_reply (sbp, NFS3ERR_BADHANDLE);
}

void
server::dispatch (svccb *sbp)
{
  if (condemned) {
    error_reply (sbp, NFS3ERR_STALE);
    return;
  }
  if (locked ()) {
    waitq.push_back (sbp);
    return;
  }
  assert (nfsc);
  if (!authok (sbp))
    return;

  if (sfsctl_intercept (this, srvno, sbp))
    return;

#ifdef USE_UVFS
  if (uvfs)
    uvfsdispatch (sbp);
  else
#endif /* USE_UVFS */
    nfs3dispatch (sbp);
}

bool
server::authok (svccb *sbp)
{
  const authunix_parms *aup = sbp->getaup ();
  if (!aup) {
    sbp->reject (AUTH_REJECTEDCRED);
    return false;
  }
  const sfs_aid aid = aup2aid (aup);
  if (aid == sfsaid_sfs || aid == sfsaid_nobody)
    return true;
  if (authnos[aid])
    return true;
  if (sbp->proc () == NFSPROC3_GETATTR
      && sbp->template getarg<nfs_fh3> ()->data == fsinfo.nfs->v3->root.data)
    return true;
  ptr<userauth> uap = authpending[aid];
  if (!uap) {
    authpending.insert (aid, New refcounted<userauth> (aid, mkref (this)));
    uap = authpending[aid];
  }
  uap->pushreq (sbp);
  return false;
}

bool
server::restart (svccb *sbp, clnt_stat err)
{
  if (condemned)
    error_reply (sbp, NFS3ERR_STALE);
  else if (err == RPC_CANTSEND || err == RPC_CANTRECV)
    waitq.push_back (sbp);
  else
    return false;
  return true;
}

void
server::sendreply (svccb *sbp, void *res, u_int32_t sno)
{
  if (!uvfs) {
    fh3trans fht (fh3trans::ENCODE, fhkey, sno);
    nfs3_exp_disable (sbp->proc (), res);
    if (!nfs3_transres (fht, res, sbp->proc ())) {
      warn ("sendreply: nfs3_transres encode failed (err %d)\n", fht.err);
      nfs3_err (sbp, nfsstat3 (fht.err));
      return;
    }
  }
#ifdef USE_UVFS
  else {
    uvfstrans fht (uvfstrans::TOUVFS, &ustate, sno);
    if (!uvfs_transres (fht, res, sbp->proc ())) {
      warn ("sendreply: uvfs_transres encode failed (err %d)\n", fht.error);
      uvfs_err (sbp, fht.error);
      return;
    }
  }
#endif /* USE_UVFS */
  sbp->reply (res);
}

void
server::getreply (time_t rqtime, svccb *sbp, void *res, clnt_stat err)
{
  auto_xdr_delete axd (ex_nfs_program_3.tbl[sbp->proc ()].xdr_res, res);
  if (!err) {
#ifndef NO_ACACHE
    if (!uvfs) {
      xattrvec xv;
      nfs3_getxattr (&xv, sbp->proc (), sbp->getvoidarg (), res);
      for (xattr *x = xv.base (); x < xv.lim (); x++) {
	if (x->fattr)
	  x->fattr->expire += rqtime;
	ac.attr_enter (*x->fh, x->fattr, x->wattr);

	if (sbp->proc () == NFSPROC3_ACCESS) {
	  ex_access3res *ares = static_cast<ex_access3res *> (res);
	  access3args *a = sbp->template getarg<access3args> ();
	  if (ares->status)
	    ac.flush_access (a->object, aup2aid (sbp->getaup ()));
	  else
	    ac.access_enter (a->object, aup2aid (sbp->getaup ()),
			      a->access, ares->resok->access);
	}
      }
    }
#endif /* !NO_ACACHE */
    if (sbp->proc () == ex_NFSPROC3_LOOKUP
	&& *static_cast<nfsstat3 *> (res) == NFS3_OK
	&& !static_cast<ex_lookup3res *> (res)->resok->obj_attributes.present)
      /* Many NFS clients have a kernel bug that causes unpredicatble
       * behavior after a missing obj_attributres on a lookup.  We
       * need to make sure we work around it in sfs. */
      panic ("lookup with no obj_attributes\n");
    sendreply (sbp, res, srvno);
  }
  else if (!restart (sbp, err))
    sbp->reject (SYSTEM_ERR);
}

void
server::nfs3dispatch (svccb *sbp)
{
#ifndef NO_ACACHE
  if (sbp->proc () == NFSPROC3_GETATTR) {
    const ex_fattr3 *f = ac.attr_lookup (*sbp->template getarg<nfs_fh3> ());
    if (f) {
      getattr3res res (NFS3_OK);
      *res.attributes = *reinterpret_cast<const fattr3 *> (f);
      sbp->reply (&res);
      return;
    }
  }
  else if (sbp->proc () == NFSPROC3_ACCESS) {
    access3args *a = sbp->template getarg<access3args> ();
    int32_t perm = ac.access_lookup (a->object, aup2aid (sbp->getaup ()),
				      a->access);
    if (perm > 0) {
      access3res res (NFS3_OK);
      res.resok->obj_attributes.set_present (true);
      *res.resok->obj_attributes.attributes
	= *reinterpret_cast<const fattr3 *> (ac.attr_lookup (a->object));
      res.resok->access = perm;
      sbp->reply (&res);
      return;
    }
  }
#endif /* !NO_ACACHE */

  void *res = ex_nfs_program_3.tbl[sbp->proc ()].alloc_res ();
  if (sbp->proc () == ex_NFSPROC3_LOOKUP)
    nfsc->call (sbp->proc (), sbp->getvoidarg (), res,
		wrap (mkref (this), &server::lookupcb_1, timenow, sbp,
		      static_cast<ex_lookup3res *> (res)),
		authnos [aup2aid (sbp->getaup ())]);
  else
    nfsc->call (sbp->proc (), sbp->getvoidarg (), res,
		wrap (mkref (this), &server::getreply, timenow, sbp, res),
		authnos [aup2aid (sbp->getaup ())]);
}

void
server::lookupcb_1 (time_t rqtime, svccb *sbp,
		    ex_lookup3res *lres, clnt_stat err)
{
  if (err || lres->status || lres->resok->obj_attributes.present) {
    getreply (rqtime, sbp, lres, err);
    return;
  }
  if (const ex_fattr3 *f = ac.attr_lookup (lres->resok->object)) {
    lres->resok->obj_attributes.set_present (true);
    *lres->resok->obj_attributes.attributes = *f;
    lres->resok->obj_attributes.attributes->expire -= rqtime;
    getreply (rqtime, sbp, lres, err);
    return;
  }
  ex_getattr3res *ares = New ex_getattr3res;
  nfsc->call (ex_NFSPROC3_GETATTR, &lres->resok->object, ares,
	      wrap (mkref (this), &server::lookupcb_2,
		    rqtime, sbp, lres, ares),
	      authnos [aup2aid (sbp->getaup ())]);
}

void
server::lookupcb_2 (time_t rqtime, svccb *sbp,
		    ex_lookup3res *lres, ex_getattr3res *ares,
		    clnt_stat err)
{
  if (!err) {
    if (ares->status)
      lres->set_status (ares->status);
    else  {
      lres->resok->obj_attributes.set_present (true);
      *lres->resok->obj_attributes.attributes = *ares->attributes;
    }
  }
  delete ares;
  getreply (rqtime, sbp, lres, err);
}

#ifdef USE_UVFS
static void
uvfsreaddirreply (ptr<server> sp, svccb *sbp, ex_readdir3res *nfsres, 
		  clnt_stat err)
{
  auto_xdr_delete axd (xdr_ex_readdir3res, nfsres);

  if (err) {
    if (!sp->restart (sbp, err))
      sbp->reject (SYSTEM_ERR);
    return;
  }

  uvfs_readdirres uvfsres (nfsres->status);
  if (nfsres->status) {
    sbp->reply (&uvfsres);
    return;
  }

  entry3 *e = nfsres->resok->reply.entries;
  int i = 0;
  while (e) {
    e = e->nextentry;
    i++;
  }

  uvfsres.reply->entries.setsize(i);
  e = nfsres->resok->reply.entries;
  for (int j = 0; j < i; j++) {
    uvfsres.reply->entries[j].fileid = e->fileid;
    uvfsres.reply->entries[j].type = 8; // don't think it matters (DT_REG)
    uvfsres.reply->entries[j].cookie = e->cookie;
    uvfsres.reply->entries[j].name = e->name;
    e = e->nextentry;
  }
  uvfsres.reply->eof = nfsres->resok->reply.eof;

  sp->sendreply (sbp, &uvfsres, sp->srvno);
}

void
server::uvfsreaddir (svccb *sbp)
{
  uvfs_readdirargs *uvfsarg = sbp->template getarg<uvfs_readdirargs> ();

  readdir3args nfsarg;
  nfsarg.dir.data = uvfsarg->dir.data;
  nfsarg.cookie = uvfsarg->cookie;
  nfsarg.count = uvfsarg->count;

  ex_readdir3res *nfsres = New ex_readdir3res (NFS3_OK);
  nfsc->call (NFSPROC3_READDIR, &nfsarg, nfsres,
	      wrap (uvfsreaddirreply, mkref (this), sbp, nfsres),
	      authnos [aup2aid (sbp->getaup ())]);
}

void
server::uvfsdispatch (svccb *sbp)
{
  switch (sbp->proc ()) {
  case UVFSPROC_READDIR:
    uvfsreaddir (sbp);
    return;
  case UVFSPROC_OPEN:
  case UVFSPROC_CLOSE:
  case UVFSPROC_INACTIVE:
    sbp->replyref (uvfsstat (0));
    return;
  case UVFSPROC_RECLAIM:
    warn << "Reclaiming fh\n";
    ustate.reclaim (sbp->template getarg<uvfs_fh> ()->data);
    sbp->replyref (uvfsstat (0));
    return;
  }

  void *res = ex_nfs_program_3.tbl[sbp->proc ()].alloc_res ();
  if (sbp->proc () == UVFSPROC_LOOKUP)
    nfsc->call (sbp->proc (), sbp->getvoidarg (), res,
		wrap (mkref (this), &server::lookupcb_1, timenow, sbp,
		      static_cast<ex_lookup3res *> (res)),
		authnos [aup2aid (sbp->getaup ())]);
  else
    nfsc->call (sbp->proc (), sbp->getvoidarg (), res,
		wrap (mkref (this), &server::getreply, timenow, sbp, res),
		authnos [aup2aid (sbp->getaup ())]);
}
#endif /* USE_UVFS */

void
server::cbdispatch (svccb *sbp)
{
  if (!sbp) {
    if (!condemned) {
      lock ();
      authnos.clear ();
      authpending.clear ();
      ac.flush_attr ();
      warn << "EOF from " << path << "\n";
      assert (!recon_tmo);
      reconnect ();
    }
    return;
  }

  if (sbp->prog () == SFSCB_PROGRAM)
    switch (sbp->proc ()) {
    case SFSCBPROC_NULL:
      sbp->reply (NULL);
      break;
    default:
      sbp->reject (PROC_UNAVAIL);
      break;
    }
  else if (sbp->prog () == ex_NFSCB_PROGRAM)
    switch (sbp->proc ()) {
    case ex_NFSCBPROC3_NULL:
      sbp->reply (NULL);
      break;
    case ex_NFSCBPROC3_INVALIDATE:
      {
	ex_invalidate3args *xa = sbp->template getarg<ex_invalidate3args> ();
	if (!uvfs) {
	  ex_fattr3 *a = NULL;
	  if (xa->attributes.present && xa->attributes.attributes->expire) {
	    a = xa->attributes.attributes.addr ();
	    a->expire += timenow;
	  }
	  ac.attr_enter (xa->handle, a, NULL);
	}
#ifdef USE_UVFS
	else if (uvfsstate::fhpair *fp = ustate.ntou_tab (xa->handle.data,
							  srvno)) {
	  uvfs_invalidateargs args;
	  args.handle.data = fp->ufh;
	  args.attributes = xa->attributes;
	  sfsc->call (UVFSCBPROC_INVALIDATE, &args, NULL, aclnt_cb_null);
	}
#endif /* USE_UVFS */
	sbp->reply (NULL);
	break;
      }
    default:
      sbp->reject (PROC_UNAVAIL);
      break;
    }
  else
    panic ("server::cbdispatch: prog %d\n", sbp->prog ());
}

u_int32_t
server::gensrvno ()
{
  while (srvno_next < srvno_min || srvnotab[srvno_next])
    srvno_next = (srvno_next + 1) & fh3trans::srvno_mask;
  return srvno_next;
}

server::server (const str &p, const sfs_connectinfo &ci,
		const sfs_servinfo &servinfo)
  : recon_tmo (NULL), recon_backoff (1), lock_flag (true), condemned (false),
    srvno (gensrvno ()), path (p), carg (ci), servinfo (servinfo), seqno (0)
{
  refcount_inc ();

  authinfo.type = SFS_AUTHINFO;
  authinfo.service = SFS_SFS;
  authinfo.name = carg.name;
  authinfo.hostid = carg.hostid;

  srvnotab.insert (this);
  pathtab.insert (this);
}

server::~server ()
{
  srvnotab.remove (this);
  pathtab.remove (this);
  while (!waitq.empty ())
    error_reply (waitq.pop_front (), NFS3ERR_STALE);
  warn << path << " deleted\n";
}

bool
server::getserver (ptr<server> *spp, u_int32_t *srvnop, svccb *sbp)
{
  assert (sbp->proc () != NFSPROC3_NULL);
  *spp = NULL;
  if (!uvfs) {
    fh3trans fht (fh3trans::DECODE, fhkey);
    if (!nfs3_transarg (fht, sbp->getvoidarg (), sbp->proc ()))
      return false;
    *srvnop = fht.srvno;
    if (fht.srvno < srvno_min)
      return true;
    if (server *s = srvnotab[fht.srvno]) {
      *spp = mkref (s);
      return true;
    }
    return false;
  }
#ifdef USE_UVFS
  else {
    /* All UVFS args except NULL start with a file handle */
    uvfstrans fht (uvfstrans::FROMUVFS, &ustate);
    if (!uvfs_transarg (fht, sbp->getvoidarg (), sbp->proc ()))
      return false;
    *srvnop = fht.srvno;
    if (server *s = srvnotab[fht.srvno]) {
      *spp = mkref (s);
      return true;
    }
    return false;
  }
#endif /* USE_UVFS */
}

void
server::mapuser (sfs_aid aid, u_int32_t authno)
{
  ptr<userauth> ua = authpending[aid];
  authpending.remove (aid);
  auto_auth aa (authuint_create (authno));
  authnos.insert (aid, aa);
}

void
server::doflushuser (sfs_aid aid, server *sp)
{
  if (AUTH *a = sp->authnos[aid]) {
    if (sp->sfsc) {
      u_int32_t authno = authuint_getval (a);
      sp->sfsc->call (SFSPROC_LOGOUT, &authno, NULL, aclnt_cb_null);
    }
    sp->authnos.remove (aid);
  }
  else if (ptr<userauth> uap = sp->authpending[aid]) {
    sp->authpending.remove (aid);
    uap->abort ();
  }
  sp->ac.flush_access (aid);
}

void
server::flushuser (sfs_aid aid)
{
  srvnotab.traverse (wrap (&server::doflushuser, aid));
}

static void
cdmountres (ref<server> s, svccb *sbp, const sfs_fsinfo *fsip, int err)
{
  s->unlock ();
  if (err) {
    sbp->replyref (sfscd_mountres (err));
    s->destroy ();
    return;
  }

  s->fsinfo = *fsip;
  sfscd_mountres res (0);
  // res.reply->mntflags = NMOPT_NOAC | NMOPT_NFS3 | NMOPT_SOFT;
  res.reply->mntflags = NMOPT_NFS3 | NMOPT_SOFT;
  // res.reply->mntflags = NMOPT_RDPLUS | NMOPT_NFS3 | NMOPT_SOFT;

#ifdef USE_UVFS
  if (uvfsfd >= 0) {
    uvfstrans fht (uvfstrans::TOUVFS, &ustate, s->srvno);
    res.reply->fh = s->fsinfo.nfs->v3->root.data;
    if (!ustate.translate (fht, res.reply->fh)) {
      warn ("could not translate root file handle\n");
      res.set_err (EINVAL);
      sbp->reply (&res);
      s->destroy ();
      return;
    }
    cdx->sendfd (uvfsfd, false);
    sbp->reply (&res);
    return;
  }
#endif /* USE_UVFS */

  fh3trans fht (fh3trans::ENCODE, fhkey, s->srvno);
  if (!rpc_traverse (fht, s->fsinfo.nfs->v3->root)) {
    warn ("could not translate root file handle\n");
    res.set_err (EINVAL);
    sbp->reply (&res);
    s->destroy ();
    return;
  }

  res.reply->fh = s->fsinfo.nfs->v3->root.data;
  s->fsinfo = *fsip;		// Reset to original file handle for reconnect
  cdx->sendfd (nfsfd, false);
  sbp->reply (&res);
  /* Things are looking good */
}

void
server::alloc (svccb *sbp, int fd)
{
  if (fd < 0) {
    sbp->replyref (sfscd_mountres (EBADF));
    return;
  }

  sfscd_mountarg *cok = sbp->template getarg<sfscd_mountarg> ();
  if (cok->cres.servinfo.prog != ex_NFS_PROGRAM
      || cok->cres.servinfo.vers != ex_NFS_V3) {
    sbp->replyref (sfscd_mountres (EPROTONOSUPPORT));
    return;
  }
  
  sfs_hash hostid;
  if (!sfs_mkhostid (&hostid, cok->cres.servinfo.host)) {
    sbp->replyref (sfscd_mountres (EINVAL));
    return;
  }
  const str path (cok->cres.servinfo.host.hostname << ":"
		  << armor32 (hostid.base (), hostid.size ()));
  if (pathtab[path]) {
    sbp->replyref (sfscd_mountres (EEXIST));
    return;
  }

  ref<server> s = New refcounted<server> (path, cok->carg, cok->cres.servinfo);
  vNew srvcon (s, fd, cok->cres, wrap (cdmountres, s, sbp));
}

void
server::reconnect ()
{
  recon_tmo = NULL;
  vNew srvcon (mkref (this), wrap (this, &server::reconnectcb));
}

void
server::reconnectcb (const sfs_fsinfo *fsip, int err)
{
  if (condemned) {
    unlock ();
    return;
  }
  if (err == EPROTONOSUPPORT || err == ENOENT) {
    warn << "reconnect " << path
	 << (err == EPROTONOSUPPORT ? ": protocol changed"
	     : ": server name or key changed\n");
    cdc->call (SFSCDCBPROC_DELFS, &path, NULL, aclnt_cb_null);
    condemn ();
    unlock ();
    return;
  }
  if (err) {
    // warn << "reconnect " << path << ": " << strerror (errno) << "\n";
    if (recon_backoff < 64)
      recon_backoff <<= 1;
    cdc->call (SFSCDCBPROC_HIDEFS, &path, NULL, aclnt_cb_null);
    recon_tmo = timecb (time (NULL) + recon_backoff,
			wrap (this, &server::reconnect));
    return;
  }
  if (fsip->nfs->v3->root.data != fsinfo.nfs->v3->root.data) {
    warn << "reconnect " << path << ": " << "root handle changed\n";
    cdc->call (SFSCDCBPROC_DELFS, &path, NULL, aclnt_cb_null);
    condemn ();
  }
  else {
    recon_backoff = 1;
    cdc->call (SFSCDCBPROC_SHOWFS, &path, NULL, aclnt_cb_null);
    warn << "reconnected to " << path << "\n";
  }
  unlock ();
}

server *
server::lookup (const str &path)
{
  return pathtab[path];
}

void
server::condemn ()
{
  if (!condemned) {
    condemned = true;
    if (recon_tmo) {
      timecb_remove (recon_tmo);
      recon_tmo = NULL;
      unlock ();
    }
    while (!waitq.empty ())
      error_reply (waitq.pop_front (), NFS3ERR_STALE);
  }
}

void
server::destroy ()
{
  condemn ();
  refcount_dec ();
}

void
server::unlock ()
{
  assert (lock_flag);
  vector<svccb *> wq;
  swap (wq, waitq);
  lock_flag = false;
  while (!wq.empty ())
    dispatch (wq.pop_front ());
}
